{*********************************************************}
{     TExcel Component Demo 2.0 for Delphi 1.0/2.0        }
{                                                         }
{     Copyright (c) 1996, 1997 by                         }
{                                                         }
{       Stefan Hoffmeister           (Advanced version)   }
{       Stefan.Hoffmeister@Uni-Passau.de                  }
{   and                                                   }
{       Tibor F. Liska                (Basic version)     }
{       Tel/Fax:    00-36-1-165-2019                      }
{       Office:     00-36-1-269-8284                      }
{       E-mail: liska@sztaki.hu                           }
{                                                         }
{*********************************************************}
unit ExcelTop;

interface

uses
  Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Spin, Buttons, Excels, DdeMan, TabNotBk{$IFDEF Win32}, ComCtrls {$ENDIF};

type
  TForm1 = class(TForm)
    cmClose: TBitBtn;
    Timer1: TTimer;
    TabbedNotebook1: TTabbedNotebook;
    cmTable: TBitBtn;
    Label1: TLabel;
    SpinEdit1: TSpinEdit;
    SpinEdit2: TSpinEdit;
    Label2: TLabel;
    Label3: TLabel;
    Edit1: TEdit;
    SpinEdit5: TSpinEdit;
    SpinEdit4: TSpinEdit;
    SpinEdit3: TSpinEdit;
    cmRequest: TBitBtn;
    cmRun: TBitBtn;
    ComboBox3: TComboBox;
    Label4: TLabel;
    cmCommand: TBitBtn;
    ComboBox1: TComboBox;
    Label6: TLabel;
    RequestMemo: TMemo;
    ComboBox4: TComboBox;
    Label7: TLabel;
    SpecialRequestEdit: TEdit;
    cmRequestSelection: TBitBtn;
    SelectionLabel: TLabel;
    CheckBox1: TCheckBox;
    CheckBox2: TCheckBox;
    cmPutData: TBitBtn;
    RadioGroup1: TRadioGroup;
    Edit2: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Timer1Timer(Sender: TObject);
    procedure ExcelOpen (Sender: TObject);
    procedure ExcelClose(Sender: TObject);
    procedure cmCloseClick(Sender: TObject);
    procedure cmTableClick(Sender: TObject);
    procedure cmCommandClick(Sender: TObject);
    procedure cmRequestClick(Sender: TObject);
    procedure cmRunClick(Sender: TObject);
    procedure cmRequestSelectionClick(Sender: TObject);
    procedure ComboBox4Change(Sender: TObject);
    procedure CheckBox1Click(Sender: TObject);
    procedure CheckBox2Click(Sender: TObject);
    procedure RadioGroup1Click(Sender: TObject);
    procedure cmPutDataClick(Sender: TObject);
  private
    MacroFile : TFileName;
{$IFNDEF INSTALLED}
    Excel: TAdvExcel;
{$ENDIF}
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.FormCreate(Sender: TObject);
begin
{$IFNDEF INSTALLED}
  Excel := TAdvExcel.Create(Self);
{$ENDIF}
  ComboBox1.ItemIndex := -1;
  ComboBox3.ItemIndex := -1;
  ComboBox4.ItemIndex := -1;
  RadioGroup1.ItemIndex := ord(Excel.MoveActiveCell);
  MacroFile := ExtractFilePath(ParamStr(0))+'TExcel.xls';
  if not FileExists(Macrofile) then
    MacroFile := '';
  Timer1.Enabled := True;          { Delayed Connect }
  Screen.Cursor := crHourGlass;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Excel.CloseMacroFile;
  Excel.OnClose := nil;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
  ExcelPath : TFileName;
begin
  Timer1.Enabled := False;
  try
    Excel.OnOpen  := ExcelOpen;
    Excel.OnClose := ExcelClose;
    try
      Excel.Connect;   { Same as Excel.Connected := True; }
    except
    ExcelPath := '';
      if InputQuery('Excel not found', 'Specify Excel path',
                                        ExcelPath) then
      begin
        Excel.ExeName := ExcelPath;
        Excel.Connect;
      end
      else
        raise;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TForm1.ExcelOpen(Sender: TObject);
begin
  cmTable  .Enabled := True;
  cmCommand.Enabled := True;
  cmRun    .Enabled := MacroFile <> '';
  cmRequest.Enabled := True;
  cmRequestSelection.Enabled := true;
  cmPutData.Enabled := True;
end;

procedure TForm1.ExcelClose(Sender: TObject);
begin
  cmTable  .Enabled := False;
  cmCommand.Enabled := False;
  cmRun    .Enabled := False;
  cmRequest.Enabled := False;
  cmRequestSelection.Enabled := False;
  cmPutData.Enabled := False;
  ShowMessage('Excel connection closed');
end;

procedure TForm1.cmCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TForm1.cmTableClick(Sender: TObject);
begin
  Excel.StartTable;        { Create new workbook }
  Excel.PutStrAt(SpinEdit1.Value, SpinEdit2.Value, Edit1.Text);
  Excel.PutIntAt(SpinEdit3.Value, SpinEdit4.Value, SpinEdit5.Value);
  Excel.EndTable;          { Show excel table }
end;

procedure TForm1.cmCommandClick(Sender: TObject);
begin
  Excel.Exec(ComboBox1.Text);
end;

procedure TForm1.cmRequestClick(Sender: TObject);
begin
  RequestMemo.Clear;
  RequestMemo.Lines.Add(Excel.Request(SpecialRequestEdit.Text));
end;

procedure TForm1.cmRunClick(Sender: TObject);
begin
  Excel.OpenMacroFile(MacroFile);
  Excel.RunMacro(ComboBox3.Text);
end;

procedure TForm1.cmRequestSelectionClick(Sender: TObject);
begin
  SelectionLabel.Caption := Excel.Selection;
end;

procedure TForm1.ComboBox4Change(Sender: TObject);
  var
    Result : TStringList;
begin
  RequestMemo.Clear;
  case ComboBox4.ItemIndex of
    0 : Result := Excel.Topics;
    1 : Result := Excel.Protocols;
    2 : Result := Excel.Formats;
    3 : Result := Excel.EditEnvItems;
  else
        Result := nil;
  end;
  if Result <> nil then
    RequestMemo.Lines.Assign(Result);
end;

procedure TForm1.CheckBox1Click(Sender: TObject);
begin
  with Excel do
    if CheckBox1.Checked then
      EchoOn
    else
      EchoOff;
end;

procedure TForm1.CheckBox2Click(Sender: TObject);
begin
  with Excel do
    if CheckBox2.Checked then
      EnableInput
    else
      DisableInput;
end;

procedure TForm1.RadioGroup1Click(Sender: TObject);
begin
  Excel.MoveActiveCell := TCellDir(RadioGroup1.ItemIndex);
end;

procedure TForm1.cmPutDataClick(Sender: TObject);
begin
  Excel.PutData( [Edit2.Text] );
(*
  { Extended example of usage: }
  Excel.PutDataAt(0, 0, [1, 2.2, 'Hello!', '', 1], dirRight);
*)
end;

end.
